/*
  media-filesystem library
  tridge@samba.org, January 2001
  released under the Gnu GPL v2
*/

#include <limits.h>

#include "mfs.h"

static union mfs_combined_superblock super;
static struct mfs_zone_info zone_info[MAX_ZONES];
static int num_zones = 0;
static int is64bit = 0;

static int total_inodes = 0;

#define MAKE_FIELD_ACCESSOR(t32bit, t64bit, basetypename, fieldtype, field) \
static fieldtype get_##basetypename##_##field(void* ptr) {                  \
	return is64bit ? ((struct t64bit*)ptr)->field                           \
	               : ((struct t32bit*)ptr)->field; }

#define MAKE_SUPER_ACCESSOR(type, field) \
MAKE_FIELD_ACCESSOR(mfs_32bit_superblock, mfs_64bit_superblock, super, type, field)

#define MAKE_ZONE_ACCESSOR(type, field) \
MAKE_FIELD_ACCESSOR(mfs_32bit_zone_map, mfs_64bit_zone_map, zonemap, type, field);

MAKE_SUPER_ACCESSOR(u64, total_sectors)
MAKE_SUPER_ACCESSOR(u64, zonemap_ptr)
MAKE_SUPER_ACCESSOR(u64, zonemap_size)
static int get_super_size() {
	return is64bit ? sizeof(struct mfs_64bit_superblock)
	               : sizeof(struct mfs_32bit_superblock);
}
static void vstream_byte_swap_super(union mfs_combined_superblock* super) {
	if (is64bit) {
		vstream_byte_swap(super, "i9 b128 i1 q10 i7");
	} else {
		vstream_byte_swap(super, "i9 b128 i17");
	}
}

MAKE_ZONE_ACCESSOR(u64, sector)
MAKE_ZONE_ACCESSOR(u64, zone_start)
MAKE_ZONE_ACCESSOR(u64, zone_size)
MAKE_ZONE_ACCESSOR(u32, type)
MAKE_ZONE_ACCESSOR(u32, per_chunk)
MAKE_ZONE_ACCESSOR(u64, next_zonemap_ptr)
MAKE_ZONE_ACCESSOR(u32, next_zonemap_size)
static u32* get_zonemap_crc_ptr(void* ptr) {
	return is64bit ? &((struct mfs_64bit_zone_map*)ptr)->crc
	               : &((struct mfs_32bit_zone_map*)ptr)->crc;
}
static void vstream_byte_swap_zone(void* zone) {
	if (is64bit) {
		vstream_byte_swap(zone, "q9 i14");
	} else {
		vstream_byte_swap(zone, "i18");
	}
}

u64 vstream_mfs_inode_run_start(struct mfs_inode* inode, int run_number)
{
	if (is64bit)
		return inode->u.runs_64bit[run_number].start;
	else
		return inode->u.runs_32bit[run_number].start;
}

u32 vstream_mfs_inode_run_length(struct mfs_inode* inode, int run_number)
{
	if (is64bit)
		return inode->u.runs_64bit[run_number].len;
	else
		return inode->u.runs_32bit[run_number].len;
}

unsigned vstream_fsid_hash(unsigned fsid, unsigned size)
{
	return fsid*67289 % size;
}

// returns 1 on failure
int load_super(void)
{
	vstream_mfs_read_partial(&super, 0, sizeof(super));

	if (super.first_short == 0x1492 || super.first_short == 0x9214) {
		if (super.first_short == 0x1492) {
			vstream_io_need_bswap(1);
			vstream_mfs_read_partial(&super, 0, sizeof(super));
		}
		if (vstream_partition_parse())
			return 1;
	}

	switch (super.header.magic) {
		case 0xebbafeed: /* normal s3 tivo access */
		case 0xedfebaeb:   /* little endian */
			is64bit = 1;
		case 0xabbafeed: /* normal tivo access */
		case 0xedfebaab:   /* little endian */
			break;

		case 0xbaebedfe:
		case 0xfeedebba:   /* little endian */
			is64bit = 1;
		case 0xbaabedfe:
		case 0xfeedabba:   /* little endian */
			vstream_io_need_bswap(1);
			vstream_mfs_read_partial(&super, 0, sizeof(super));
			break;

	default:
		vstream_error("Not a consistent TiVo super block! (magic=0x%08x)\n",
		              super.header.magic);
		return 1;
	}

	int superblock_size = get_super_size();
	vstream_check_crc((void *)&super, superblock_size, &super.header.crc);
	
	vstream_byte_swap_super(&super);

	if (super.header.magic != 0xabbafeed &&
	    super.header.magic != 0xebbafeed) {
		vstream_error("Failed to byte swap correctly\n");
		return 1;
	}

	if (vstream_partition_total_size() && 
	    vstream_partition_total_size() != get_super_total_sectors(&super)) {
		vstream_error("WARNING: total sectors doesn't match (total=%d sb=%d)\n",
		              vstream_partition_total_size(),
		              get_super_total_sectors(&super));
	}

	return 0;
}

/* load the mfs zones - currently we only use the inode
   zone but might as well load the lot */
int load_zones(void)
{
	u64 next = get_super_zonemap_ptr(&super);
	u32 map_size = get_super_zonemap_size(&super);
	total_inodes = 0;

	void* zone_buffer = NULL;
	while (next) {
		zone_buffer = realloc(zone_buffer, map_size*SECTOR_SIZE);
		vserver_vstream_read_sectors(zone_buffer, next, map_size);
		vstream_check_crc(zone_buffer, map_size*SECTOR_SIZE,
		                  get_zonemap_crc_ptr(zone_buffer));
		vstream_byte_swap_zone(zone_buffer);
		if (next != get_zonemap_sector(zone_buffer)) {
			vstream_error("sector wrong in zone (%d %d)\n",
			              next, get_zonemap_sector(zone_buffer));
			return 1;
		}
		if (get_zonemap_type(zone_buffer) == ZONE_INODE) {
			total_inodes += get_zonemap_zone_size(zone_buffer)/2;
		}

		zone_info[num_zones].sector = get_zonemap_sector(zone_buffer);
		zone_info[num_zones].zone_start = get_zonemap_zone_start(zone_buffer);
		zone_info[num_zones].zone_size = get_zonemap_zone_size(zone_buffer);
		zone_info[num_zones].type = get_zonemap_type(zone_buffer);
		zone_info[num_zones].per_chunk = get_zonemap_per_chunk(zone_buffer);
		zone_info[num_zones].zonemap_size = map_size;

		next = get_zonemap_next_zonemap_ptr(zone_buffer);
		map_size = get_zonemap_next_zonemap_size(zone_buffer);
		num_zones++;
		if (num_zones == MAX_ZONES) {
			vstream_error("Too many zones\n");
			return 1;
		}
	}
	free(zone_buffer);
	return 0;
}

/* turn a hash into a zone sector */
static u64 zone_sector(u32 hash)
{
	int i;
	u64 start = 0;
	for (i=0;i<num_zones;i++) {
		u32 len;
		if (zone_info[i].type != ZONE_INODE)
			continue;
		len = zone_info[i].zone_size/2;
		if (hash < start + len) {
			return zone_info[i].zone_start + (hash-start)*2;
		}
		start += len;
	}
	vstream_error("Didn't find hash %x in zones!\n", hash);
	return ULLONG_MAX;
}

static void vstream_mfs_byte_swap_inode(struct mfs_inode *inode, bool from_disk)
{
	if (from_disk) {
		vstream_byte_swap(inode, "i10 b2 s1 i4");
	}
	if (inode->num_runs) {
		int i;
		if (is64bit) {
			for(i = 0; i < inode->num_runs; i++)
				vstream_byte_swap(&inode->u.runs_64bit[i], "q1 i1" );
		} else {
			for(i = 0; i < inode->num_runs; i++)
				vstream_byte_swap(&inode->u.runs_32bit[i], "i2" );
		}
	}
	if (!from_disk) {
		vstream_byte_swap(inode, "i10 b2 s1 i4");
	}
}

/* load one inode by fsid - optimised to avoid repeats */
static int vstream_mfs_load_inode(int fsid, struct mfs_inode *inode)
{
	static struct mfs_inode in;
	static u32 last_fsid;
	unsigned hash, hash1;

	if (fsid && fsid == last_fsid) {
		*inode = in;
		return 0;
	}

	hash1 = hash = vstream_fsid_hash(fsid, total_inodes);
	do {
		u64 zs = zone_sector(hash);
		if (zs == ULLONG_MAX) return 1;
		vstream_mfs_read_partial(&in, zs, sizeof(in));
		vstream_check_crc((void *)&in, sizeof(in), &in.crc);
		vstream_mfs_byte_swap_inode(&in, true);
		hash = (hash+1) % total_inodes;
	} while ((in.flags & MFS_FLAGS_CHAIN) && in.id != fsid && hash != hash1);

	if (in.id != fsid) {
		vstream_error("ERROR: Didn't find fsid=%d!\n", fsid);
		return 1;
	}

	*inode = in;
	last_fsid = fsid;

	return 0;
}

/* read count bytes from a mfs file at offset ofs,
   returning the number of bytes read 
   ofs must be on a sector boundary
*/
u32 vstream_mfs_fsid_pread(int fsid, void *buf, u64 ofs, u32 count)
{
	struct mfs_inode inode;
	int i, n;
	u64 start;
	u32 ret=0;
	u64 sec = ofs >> SECTOR_SHIFT;
	u64 size;

	if (vstream_mfs_load_inode(fsid, &inode)) return 0;

	if (inode.num_runs == 0) {
		if (ofs >= inode.size) return 0;
		ret = count;
		if (ret > inode.size-ofs) {
			ret = inode.size-ofs;
		}
		memcpy(buf, inode.u.data + ofs, ret);
		return ret;
	}

	size = inode.size;
	if (inode.units == 0x20000) {
		size <<= 17;
	}

	if (ofs > size) return 0;
	if (ofs + count > size) {
		count = size-ofs;
	}

	while (count > 0) {
		u32 n2;
		start = 0;
		u32 run_len;
		for (i=0; i<inode.num_runs; i++) {
			run_len = vstream_mfs_inode_run_length(&inode, i);
			if (sec < start + run_len) break;
			start += run_len;
		}
		if (i == inode.num_runs) return ret;
		n = (count+(SECTOR_SIZE-1))>>SECTOR_SHIFT;
		if (n > run_len-(sec-start)) {
			n = run_len-(sec-start);
		}
		n2 = n<<SECTOR_SHIFT;
		if (n2 > count) n2 = count;
		u64 run_start = vstream_mfs_inode_run_start(&inode, i);
		vstream_mfs_read_partial(buf, run_start+(sec-start), n2);

		buf += n2;
		sec += n;
		count -= n2;
		ret += n2;
	}
	return ret;
}

/* return the type of a fsid */
int vstream_mfs_fsid_type(int fsid)
{
	struct mfs_inode inode;
	if (fsid == 0) return 0;
	memset(&inode, 0, sizeof(inode));
	if (vstream_mfs_load_inode(fsid, &inode)) return 0;
	return inode.type;
}

/* return the number of bytes used by a fsid */
u64 vstream_mfs_fsid_size(int fsid)
{
	struct mfs_inode inode;

	if (fsid == 0) return 0;
	if (vstream_mfs_load_inode(fsid, &inode)) return 0;

	switch (inode.units) {
	case 0: return inode.size;
	case 0x20000: return ((u64)inode.size) << 17;
	}
	vstream_error("ERROR: fsid=%d Unknown units %d\n", fsid, inode.units);
	return inode.size;
}

/* list a mfs directory - make sure you free with vstream_mfs_dir_free() */
struct mfs_dirent *mfs_dir(int fsid, u32 *count)
{
	u32 *buf, *p;
	int n=0, i;
	int size = vstream_mfs_fsid_size(fsid);
	int dsize, dflags;
	struct mfs_dirent *ret;

	*count = 0;

	if (size < 4) return NULL;

	if (vstream_mfs_fsid_type(fsid) != MFS_TYPE_DIR) {
		vstream_error("fsid %d is not a directory\n", fsid);
		return NULL;
	}

	buf = (u32 *)malloc(size);
	vstream_mfs_fsid_pread(fsid, buf, 0, size);
	dsize = ntohl(buf[0]) >> 16;
	dflags = ntohl(buf[0]) & 0xFFFF;
	p = buf + 1;
	while ((int)(p-buf) < dsize/4) {
		u8 *s = ((char *)p)+4;
		p += s[0]/4;
		n++;
	}
	ret = malloc((n+1)*sizeof(*ret));
	p = buf + 1;
	for (i=0;i<n;i++) {
		u8 *s = ((char *)p)+4;
		ret[i].name = strdup(s+2);
		ret[i].type = s[1];
		ret[i].fsid = ntohl(p[0]);
		p += s[0]/4;
	}	
	ret[n].name = NULL;
	free(buf);
	*count = n;

	/* handle meta-directories. These are just directories which are
	   lists of other directories. All we need to do is recursively read
	   the other directories and piece together the top level directory */
	if (dflags == 0x200) {
		struct mfs_dirent *meta_dir = NULL;
		int meta_size=0;

		*count = 0;

		for (i=0;i<n;i++) {
			struct mfs_dirent *d2;
			u32 n2;
			if (ret[i].type != MFS_TYPE_DIR) {
				vstream_error("ERROR: non dir %d/%s in meta-dir %d!\n", 
				       ret[i].type, ret[i].name, fsid);
				continue;
			}
			d2 = mfs_dir(ret[i].fsid, &n2);
			if (!d2 || n2 == 0) continue;
			meta_dir = realloc(meta_dir, sizeof(ret[0])*(meta_size + n2 + 1));
			memcpy(meta_dir+meta_size, d2, n2*sizeof(ret[0]));
			meta_size += n2;
			free(d2);
		}
		vstream_mfs_dir_free(ret);
		if (meta_dir) meta_dir[meta_size].name = NULL;
		*count = meta_size;
		return meta_dir;
	}


	return ret;
}

/* free a dir from mfs_dir */
void vstream_mfs_dir_free(struct mfs_dirent *dir)
{
	int i;
	for (i=0; dir[i].name; i++) {
		free(dir[i].name);
		dir[i].name = NULL;
	}
	free(dir);
}

/* resolve a path to a fsid */
u32 vstream_mfs_resolve(char *path)
{
	char *p0, *tok, *r = NULL;
	u32 fsid;
	struct mfs_dirent *dir = NULL;

	if (path[0] != '/') {
		return atoi(path);
	}

	fsid = MFS_ROOT_FSID;
	p0 = strdup(path);
	path = p0;
	for (tok=strtok_r(path,"/", &r); tok; tok=strtok_r(NULL,"/", &r)) {
		u32 count;
		int i;
		dir = mfs_dir(fsid, &count);
		if (!dir) {
			vstream_error("resolve failed for fsid=%d\n", fsid);
			return 0;
		}
		for (i=0;i<count;i++) {
			if (strcmp(tok, dir[i].name) == 0) break;
		}
		if (i == count) {
			fsid = 0;
			goto done;
		}
		fsid = dir[i].fsid;
		if (dir[i].type != MFS_TYPE_DIR) {
			if (strtok_r(NULL, "/", &r)) {
				vstream_error("not a directory %s\n",tok);
				fsid = 0;
				goto done;
			}
			goto done;
		}
		vstream_mfs_dir_free(dir);
		dir = NULL;
	}

 done:
	if (dir) vstream_mfs_dir_free(dir);
	if (p0) free(p0);
	return fsid;
}
